package io.gitlab.jfronny.commons.serialize.gson.api.v1;

import com.google.gson.*;
import io.gitlab.jfronny.commons.ComparableVersion;
import io.gitlab.jfronny.commons.serialize.gson.impl.ComparableVersionAdapter;
import io.gitlab.jfronny.commons.serialize.gson.impl.GsonIgnoreExclusionStrategy;

import java.lang.reflect.Modifier;
import java.lang.reflect.Type;
import java.util.function.Consumer;

/**
 * Holds a common instance of the Gson object.
 * Supports registering type adapters/etc. as needed
 */
public class GsonHolder {
    private final GsonBuilder builder = new GsonBuilder()
            .registerTypeAdapter(ComparableVersion.class, new ComparableVersionAdapter())
            .excludeFieldsWithModifiers(Modifier.TRANSIENT, Modifier.PRIVATE)
            .setExclusionStrategies(new GsonIgnoreExclusionStrategy())
            .setLenient();

    private boolean clean = false;
    private Gson gson;

    public GsonHolder() {
        synchronized (GsonHolders.KNOWN_INSTANCES) {
            GsonHolders.KNOWN_INSTANCES.add(this);
            for (Consumer<GsonBuilder> modification : GsonHolders.KNOWN_MODIFICATIONS) modification.accept(builder);
        }
    }

    /**
     * Get the current gson instance or build it if needed
     * @return The Gson instance
     */
    public Gson getGson() {
        if (!clean) {
            gson = builder.create();
            clean = true;
        }
        return gson;
    }

    /**
     * Register a type adapter and mark the gson instance as unclean
     * @param type The type for which to register the adapter
     * @param typeAdapter The adapter type
     */
    public GsonHolder registerTypeAdapter(Type type, Object typeAdapter) {
        builder.registerTypeAdapter(type, typeAdapter);
        clean = false;
        return this;
    }

    /**
     * Register a type adapter factory and mark the gson instance as unclean
     * @param factory The factory to register
     */
    public GsonHolder registerTypeAdapterFactory(TypeAdapterFactory factory) {
        builder.registerTypeAdapterFactory(factory);
        clean = false;
        return this;
    }

    /**
     * Run a function on the builder for modifying it and mark the gson instance as unclean
     * @param func The function to run
     */
    public GsonHolder modifyBuilder(Consumer<GsonBuilder> func) {
        func.accept(builder);
        clean = false;
        return this;
    }
}
