package io.gitlab.jfronny.muscript.error;

/**
 * Class for storing errors with code context
 * Can be generated from a LocationalException with asPrintable
 */
public record LocationalError(String message, String line, int lineNumber, int column) {
    /**
     * Generates an error by getting the relevant data from what is available while lexing/parsing
     * @param source The complete source for the code that generated this error
     * @param character The index of the current character
     * @param message The error message
     * @return An error using the provided information
     */
    public static LocationalError create(String source, int character, String message) {
        int lineStart = source.lastIndexOf('\n', character);
        int lineEnd = source.indexOf('\n', character);
        if (lineEnd == -1) lineEnd = source.length();

        int lineIndex = lineStart > 0 ? (int) source.substring(0, lineStart).chars().filter(c -> c == '\n').count() : 0;
        int column = character - lineStart;

        return new LocationalError(message, source.substring(lineStart + 1, lineEnd), lineIndex + 1, column);
    }

    /**
     * Converts this error to a human-readable error message
     * @return A string representation
     */
    @Override
    public String toString() {
        StringBuilder builder = new StringBuilder();

        if (column > line.length())
            builder.append("Error at character ").append(column);
        else
            builder.append("Error at '").append(line.charAt(column - 1)).append("' (character ").append(column).append(")");
        builder.append(": ").append(message);

        String linePrefix = String.format("%1$6d", lineNumber) + " | ";

        builder.append('\n').append(linePrefix).append(line);
        builder.append('\n').append(" ".repeat(linePrefix.length() + (column - 1))).append("^-- Here");

        return builder.toString();
    }
}
