package io.gitlab.jfronny.commons;

import io.gitlab.jfronny.commons.serialize.Serializer;

import java.io.*;
import java.lang.reflect.Type;
import java.net.*;
import java.net.http.HttpClient;
import java.net.http.HttpRequest;
import java.net.http.HttpResponse;
import java.nio.charset.StandardCharsets;
import java.util.*;
import java.util.function.Predicate;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class HttpUtils {
    private static String userAgent = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36";
    private static final String PROXY_AUTH;
    private static final HttpClient CLIENT;

    static {
        // Enables HTTPS proxying
        System.setProperty("jdk.http.auth.tunneling.disabledSchemes", "");

        HttpClient.Builder clientBuilder = HttpClient.newBuilder().followRedirects(HttpClient.Redirect.ALWAYS);

        String host = System.getProperty("http.proxyHost");
        String port = System.getProperty("http.proxyPort");
        if (host != null && port != null) {
            if (port.matches("\\d+")) {
                clientBuilder.proxy(ProxySelector.of(new InetSocketAddress(host, Integer.parseInt(port))));
            } else {
                System.err.println("Could not parse proxy settings: Port is not a number");
            }
        }

        String name = System.getProperty("http.proxyUserName");
        String pass = System.getProperty("http.proxyUserPassword");
        if (name != null && pass != null) {
            PROXY_AUTH = "Basic " + new String(Base64.getEncoder().encode((name + ":" + pass).getBytes()));
        } else {
            PROXY_AUTH = null;
        }

        CLIENT = clientBuilder.build();
    }

    public static void setUserAgent(String hostname) {
        if (hostname == null || hostname.isEmpty()) throw new IllegalArgumentException("Hostname cannot be empty");
        HttpUtils.userAgent = hostname;
    }

    private enum Method {
        GET,
        POST,
        PUT,
        PATCH,
        DELETE
    }

    public static class Request {
        private static final Predicate<String> CURSEFORGE_API = Pattern.compile("(?:http(s)?://)?addons-ecs\\.forgesvc\\.net/api/+").asMatchPredicate();
        private final String url;
        private final HttpRequest.Builder builder;
        private Method method;
        private int sent = 0;

        public Request(Method method, String url) throws URISyntaxException {
            this.url = url.replace(" ", "%20");
            this.builder = HttpRequest.newBuilder()
                    .uri(new URI(this.url))
                    .header("User-Agent", userAgent);
            this.method = method;
        }

        public Request bearer(String token) {
            builder.header("Authorization", "Bearer " + token);

            return this;
        }

        public Request header(String name, String value) {
            builder.header(name, value);
            return this;
        }

        public Request bodyString(String string) {
            builder.header("Content-Type", "text/plain");
            builder.method(method.name(), HttpRequest.BodyPublishers.ofString(string));
            method = null;

            return this;
        }

        public Request bodyForm(String string) {
            builder.header("Content-Type", "application/x-www-form-urlencoded");
            builder.method(method.name(), HttpRequest.BodyPublishers.ofString(string));
            method = null;

            return this;
        }

        public Request bodyForm(Map<String, String> entries) {
            return bodyForm(entries.entrySet()
                    .stream()
                    .map(entry -> URLEncoder.encode(entry.getKey(), StandardCharsets.UTF_8) + '=' + URLEncoder.encode(entry.getValue(), StandardCharsets.UTF_8))
                    .collect(Collectors.joining("&")));
        }

        public Request bodyJson(String string) {
            builder.header("Content-Type", "application/json");
            builder.method(method.name(), HttpRequest.BodyPublishers.ofString(string));
            method = null;

            return this;
        }

        public Request bodySerialized(Object object) throws IOException {
            Serializer serializer = Serializer.getInstance();
            builder.header("Content-Type", serializer.getFormatMime());
            builder.method(method.name(), HttpRequest.BodyPublishers.ofString(serializer.serialize(object)));
            method = null;

            return this;
        }

        private <T> T _send(String accept, HttpResponse.BodyHandler<T> responseBodyHandler) throws IOException {
            sent++;
            if (sent > 3) throw new IOException("Attempted third redirect, stopping");
            builder.header("Accept", accept);
            if (method != null) builder.method(method.name(), HttpRequest.BodyPublishers.noBody());
            if (PROXY_AUTH != null) builder.header("Proxy-Authorization", PROXY_AUTH);

            HttpResponse<T> res;
            try {
                res = CLIENT.send(builder.build(), responseBodyHandler);
            } catch (InterruptedException e) {
                throw new IOException("Could not send request", e);
            }
            if (res.statusCode() / 100 == 2) return res.body();
            Optional<String> location = res.headers().firstValue("location");
            // Redirect
            if (location.isPresent() && (res.statusCode() == 302 || res.statusCode() == 307) && method == Method.GET) {
                try {
                    return HttpUtils.get(location.get())._send(accept, responseBodyHandler);
                } catch (URISyntaxException e) {
                    throw new IOException("Could not follow redirect", e);
                }
            }
            // CurseForge serverside error
            if (CURSEFORGE_API.test(url) && res.statusCode() >= 500 && res.statusCode() < 600) {
                try {
                    Thread.sleep(1000);
                } catch (InterruptedException e) {
                    throw new IOException("Could not sleep before resending request" + e);
                }
                return _send(accept, responseBodyHandler);
            }
            if (res.statusCode() == 404) throw new FileNotFoundException("Didn't find anything under that url (URL=" + url + ")");
            throw new IOException("Unexpected return method: " + res.statusCode() + " (URL=" + url + ")");
        }

        public void send() throws IOException {
            _send("*/*", HttpResponse.BodyHandlers.discarding());
        }

        public InputStream sendInputStream() throws IOException {
            return _send("*/*", HttpResponse.BodyHandlers.ofInputStream());
        }

        public Reader sendReader() throws IOException {
            return new InputStreamReader(sendInputStream());
        }

        public String sendString() throws IOException {
            return _send("*/*", HttpResponse.BodyHandlers.ofString());
        }

        public Stream<String> sendLines() throws IOException {
            return _send("*/*", HttpResponse.BodyHandlers.ofLines());
        }

        public <T> T sendSerialized(Type type) throws IOException {
            Serializer serializer = Serializer.getInstance();
            InputStream in = _send(serializer.getFormatMime(), HttpResponse.BodyHandlers.ofInputStream());
            try {
                return in == null ? null : serializer.deserialize(new InputStreamReader(in), type);
            } catch (Serializer.SerializeException e) {
                throw new IOException("Could not deserialize", e);
            }
        }

        private String getString(Object a) throws IOException {
            if (a instanceof InputStream s) return new String(s.readAllBytes());
            if (a instanceof String s) return s;
            if (a instanceof Stream s) return ((Stream<String>)s).collect(Collectors.joining());
            return "";
        }
    }

    public static Request get(String url) throws URISyntaxException {
        return new Request(Method.GET, url);
    }

    public static Request post(String url) throws URISyntaxException {
        return new Request(Method.POST, url);
    }

    public static Request put(String url) throws URISyntaxException {
        return new Request(Method.PUT, url);
    }

    public static Request patch(String url) throws URISyntaxException {
        return new Request(Method.PATCH, url);
    }

    public static Request delete(String url) throws URISyntaxException {
        return new Request(Method.DELETE, url);
    }
}
