package io.gitlab.jfronny.commons.throwable;

import org.jetbrains.annotations.ApiStatus;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;

import java.util.Objects;
import java.util.function.Function;
import java.util.function.Supplier;

@FunctionalInterface
public interface ThrowingSupplier<T, TEx extends Throwable> {
    T get() throws TEx;

    @Contract(value = "_ -> new", pure = true) @ApiStatus.NonExtendable
    default <V> @NotNull ThrowingSupplier<V, TEx> andThen(@NotNull ThrowingFunction<? super T, ? extends V, ? extends TEx> after) {
        Objects.requireNonNull(after);
        return () -> after.apply(this.get());
    }

    @Contract(value = "_ -> new", pure = true) @ApiStatus.NonExtendable
    default @NotNull ThrowingRunnable<TEx> andThen(@NotNull ThrowingConsumer<? super T, ? extends TEx> after) {
        Objects.requireNonNull(after);
        return () -> after.accept(this.get());
    }

    @Contract(value = "_ -> new", pure = true) @ApiStatus.NonExtendable
    default @NotNull Supplier<T> addHandler(@NotNull Function<Throwable, ? extends T> handler) {
        Objects.requireNonNull(handler);
        return () -> {
            try {
                return this.get();
            } catch (Throwable e) {
                return handler.apply(e);
            }
        };
    }

    @Contract(value = "_, _ -> new", pure = true) @ApiStatus.NonExtendable
    default @NotNull Supplier<T> addHandler(@NotNull Class<TEx> exception, @NotNull Function<TEx, ? extends T> handler) {
        Objects.requireNonNull(exception);
        Objects.requireNonNull(handler);
        return () -> {
            try {
                return this.get();
            } catch (Throwable e) {
                if (exception.isAssignableFrom(e.getClass()))
                    return handler.apply((TEx) e);
                else throw Try.runtimeException(e);
            }
        };
    }

    @Contract(value = "-> new", pure = true) @ApiStatus.NonExtendable
    default @NotNull Supplier<T> orThrow() {
        return orThrow(Try::runtimeException)::get;
    }

    @Contract(value = "_ -> new", pure = true) @ApiStatus.NonExtendable
    default <TEx1 extends Throwable> @NotNull ThrowingSupplier<T, TEx1> orThrow(@NotNull Function<Throwable, TEx1> generator) {
        Objects.requireNonNull(generator);
        return () -> {
            try {
                return this.get();
            } catch (Throwable e) {
                throw generator.apply(e);
            }
        };
    }
}
